##############################################################################
#
# Copyright (c) 2003 gocept gmbh & co. kg. All rights reserved.
#
# See also LICENSE.txt
#
##############################################################################
"""CMF link checker tool
$Id: LinkCheckerTool.py,v 1.38 2004/07/01 06:52:36 ctheune Exp $"""

# Python stuff
from types import UnicodeType, StringType
StringTypes = (UnicodeType, StringType)
from md5 import md5
from random import shuffle
import urllib
import urlparse
import sys, traceback

# Global Zope stuff
from DateTime import DateTime
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder
from Globals import InitializeClass, PersistentMapping, Persistent
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.CMFCorePermissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from AccessControl import getSecurityManager
from ZPublisher.Publish import get_module_info
from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
from ZODB.POSException import ReadConflictError

# Sibling imports
from interfaces import ILinkCheckerTool, ILinkInformation
from LinkInfo import LinkInfo
import retrievers

import zLOG

# Threading stuff
from threading import Thread
try:
    from threading import BoundedSemaphore
except ImportError:
    from threading import Semaphore
    BoundedSemaphore = Semaphore
import time
import Queue

thread_manager = []

# Link checking stuff
import protocols

retrieval_queue = Queue.Queue(0) # Make this an infinite synchronized queue
thread_stop = 0

thread_count = 0
scheduler_count = 0
pool_threads = BoundedSemaphore(value=1)
pool_scheduler = BoundedSemaphore(value=1)

last_urls = []

class TooManyThreadsRunning(Exception):
    pass

class UnknownLink(Exception):
    pass

class LinkCheckerTool(UniqueObject, ActionProviderBase, Folder):
    """A link checker tool. Encapsulating mechanisms for finding links in the
       site, testing their availability and notifying their owners."""

    __implements__ = (ILinkCheckerTool, )

    id = 'portal_linkchecker'
    meta_type = 'CMF Linkchecker Tool'

    security = ClassSecurityInfo()
    security.setPermissionDefault("Linkchecker: Run mail notifications", ("Manager",))

    plone_tool = 1

    manage_options = ( ({'label' : 'Overview', 'action' : 'manage_overview'},) +
                        ActionProviderBase.manage_options +
                        Folder.manage_options )

    _actions = [ActionInformation(id='linkchecker_object_status',
                                  title='Links',
                                  action=Expression(text='string:${object_url}/lc_object_status'),
                                  permissions=('Modify portal content',),
                                  category='object',
                                  visible=1),
                ActionInformation(id='linkchecker_member_overview',
                                 title='My Links',
                                 action=Expression(text='string:${portal_url}/lc_my_dead_links'),
                                 condition='member',
                                 category='user',
                                 visible=1),
                ActionInformation(id='linkchecker_balanced_scorecard',
                                 title='Balanced Score Card: Links',
                                 category='portal_tabs',
                                 permissions=('Manage portal',),
                                 visible=1,
                                 action=Expression(text='string:${portal_url}/lc_balanced_score_card'))]

    maximumCacheAge = 60*60     # 1 hour
    gracefulUpdateAge = 60*45   # 45 minutes
    checkingPeriod = 24   # 1 day to check all links
    defaultURIPrefix = ""       # No prefix by default

    #
    # ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = PageTemplateFile('www/explainLinkcheckerTool', globals(),
                                        __name__='manage_overview')

    security.declareProtected(ManagePortal, 'edit_configuration')
    def edit_configuration(self, maxage, graceage, defaultURIPrefix):
        """Update the configuration options."""
        self.setMaximumCacheAge(maxage)
        self.setGracefulUpdateAge(graceage)
        self.setDefaultURIPrefix(defaultURIPrefix)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_configure')

    security.declareProtected(ManagePortal, 'manage_checkLink')
    def manage_checkLink(self, id):
        """ZMI method for checking a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
                       
        link = getattr(self, id)
        self.checkLink(link)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Link+checked')

    def manage_forceCheckLink(self, url):
        """ZMI method for forcing the check of a link."""
        if not self._cache.has_key(url):
            if hasattr(self.REQUEST, 'RESPONSE'):
                self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=No+such+link')
            return
        link = getattr(self, id)
        self.__queueAddLink(link)
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checklinks?manage_tabs_message=Check+forced')

    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_checkAll(self):
        """ZMI method for checking all links."""
        self.checkAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+checked+or+queued.')
        
    security.declareProtected(ManagePortal, 'manage_checkAll')
    def manage_forceCheckAll(self):
        """ZMI method for checking all links."""
        self.forceCheckAllLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checker?manage_tabs_message=All+links+forced+into+queue.')

    security.declareProtected(ManagePortal, 'manage_findLinks')
    def manage_findLinks(self):
        """ZMI method for finding new links."""
        count = self.findLinks()
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_checklinks?manage_tabs_message=%s+links+added.+%s+links+removed.' % count) 
            
    security.declareProtected(ManagePortal, 'manage_activateWorkerThread')
    def manage_activateWorkerThread(self):
        """ZMI method for activating a new worker thread."""
        try:
            self.activateWorkerThread()
            message = "Activated worker thread."
        except TooManyThreadsRunning:
            message = "Already %s running." % thread_count
            
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checker?manage_tabs_message=%s' % message)

    security.declareProtected(ManagePortal, 'manage_activateSchedulerThread')
    def manage_activateSchedulerThread(self):
        """ZMI method for activating a new scheduler thread."""
        try:
            self.activateSchedulerThread()
            message = "Activated worker thread."
        except TooManyThreadsRunning:
            message = "Already %s running." % scheduler_count
        message = "Scheduler+activated"
        if hasattr(self.REQUEST, 'RESPONSE'):
            self.REQUEST.RESPONSE.redirect('manage_checker?manage_tabs_message=%s' % message)

    #
    # Interface methods
    #
    def __init__(self):
        self._retrievers = PersistentMapping()

        # register all global retrievers
        for type, retriever in retrievers.retrievers.items():
            self.registerRetriever(type, retriever)

    security.declarePrivate("getLinkId")
    def getLinkId(self, url):
        """The ID is the md5 sum of the link."""
        if url is None:
            url = ""
        return md5(url).hexdigest()

    security.declarePrivate("getInfoForURL")
    def getInfoForURL(self, url):
        """Get an info object for this url"""
        id = self.getLinkId(url)
        info = getattr(self, id, None)
        if info is None:
            info = LinkInfo(url, self.REQUEST)
            self._setObject(id, info)
            info = getattr(self, id)
        return info
        
    def manage_afterAdd(self, container, object):
        # The link catalog that manages links
        self.manage_addProduct['ZCatalog'].manage_addZCatalog('link_catalog', 'Link Catalog')
        self.link_catalog.addIndex('url', 'FieldIndex')
        self.link_catalog.addIndex('state', 'KeywordIndex')
        self.link_catalog.addIndex('objects', 'KeywordIndex')
        self.link_catalog.addIndex('lastupdate', 'DateIndex')
       
        self.link_catalog.addColumn("lastcheck", None)
        self.link_catalog.addColumn("lastupdate", None)
        self.link_catalog.addColumn("oldstate", None)
        self.link_catalog.addColumn("objects", [])
        self.link_catalog.addColumn("reason", "")
        self.link_catalog.addColumn("state", "")
        self.link_catalog.addColumn("url", "")

    ##########################################################################
    ## UID methods
    security.declarePrivate('getUIDForObject')
    def getUIDForObject(self, object):
        if object is None:
            return None
        uid = None
        try:
           
            uid = object.aq_base.portal_linkchecker_uid
        except AttributeError:
            uid_a = object.getId()
            uid_b = ""
            uid = lambda x,y: "%s-%s" % (x, y)
            while uid(uid_a, uid_b) in self.portal_catalog.\
                                Indexes['portal_linkchecker_uid'].uniqueValues():
                uid_b = time.time()

            object.portal_linkchecker_uid = uid(uid_a, uid_b)
            self.portal_catalog.reindexObject(object)
            uid = object.portal_linkchecker_uid
        return uid

    security.declarePrivate('getObjectForUID')
    def getObjectForUID(self, uid):
        results = self.portal_catalog(portal_linkchecker_uid=uid)
        if not results:
                return None
        else:
                return results[0].getObject()

    ## registry functions for link information retrieval
    security.declareProtected(ManagePortal, 'registerRetriever')
    def registerRetriever(self, portal_type, function_ref):
        assert not self._retrievers.has_key(portal_type), \
            "Duplicate portal_type registration (%s)" % portal_type
        self._retrievers[portal_type] = function_ref

    def getRetrievableTypes(self):
        return self._retrievers.keys()

    def getRetriever(self, type):
        return self._retrievers[type]
        
    security.declareProtected("View", 'checkLink')
    def checkLink(self, link):
        """Run a check on the link."""

        now = time.time()
        if link.lastcheck:
            age = link.lastcheck.millis() / 1000
        else:
            age = 0
        
        # Cache invalid? Requeue an update and return "unknown"
        if now - age > self.maximumCacheAge:
            self.__queueAddLink(link)
            return "grey"

        # Update required?
        if now - age > self.gracefulUpdateAge:
            self.__queueAddLink(link)
            return "grey"
        
        return link.state

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def checkAllLinks(self):
        for link in self.objectValues(["Link Information"]):
            self.checkLink(link)

    security.declareProtected(ManagePortal, 'checkAllLinks')
    def forceCheckAllLinks(self):
        for link in self.objectValues(["Link Information"]):
            self.__queueAddLink(link)

    security.declarePrivate('clearLinks')
    def clearLinks(self):
        """Removes all links that aren't referenced anymore."""
        to_delete = []
        for x in self.objectValues(["Link Information"]):
            if len(x.objects) == 0:
                to_delete.append(x.getId())
        self.manage_delObjects(to_delete)

    security.declareProtected(ManagePortal, 'findLinks')
    def findLinks(self):
        """Find new links in the site.

        Finds new links from all objects that are retrievable and
        store them in the cache. No link check is done here."""
        count = 0
        links = []

        # Clear all uid references
        for x in self.objectValues(["Link Information"]):
            x.clearObjects()

        # gather all objects that are of a type we can check for links
        for key in self.getRetrievableTypes():
            objects = self.portal_catalog(portal_type=key)
            for ob in objects:
                ob = ob.getObject()     
                if ob is None:  # Maybe the catalog isn't up to date
                    continue
                self.findLinksForObject(ob)

        # Delete old references
        self.clearLinks()

    security.declareProtected('View', 'findLinksForObject')
    def findLinksForObject(self, object):
        
        # Check for View-Permission on context object (You need this in addition
        # to the View-Permission for the linkchecker.
        sm = getSecurityManager()
        if not sm.checkPermission("View", object):
            raise "Unauthorized", "Can't check object for links"   

        # Delete old references for this object
        links = self.link_catalog(objects=[self.getUIDForObject(object)])
        links = [ x.getObject() for x in links ]
        for x in links:
             x.removeObject(object)
        # Update links
        type = object.Type()
        if not type in self.getRetrievableTypes():
            return
        
        retriever = self.getRetriever(type)
        retriever(self, object)

        # Remove stale links
        self.clearLinks()

    security.declareProtected(ManagePortal, 'showLinks')
    def showLinks(self):
        """Show a grouped list of link data."""
        data = self._cache.values()

        # group them all
        grouped = {}
        for x in data:
            group = x.getGroupId()
            if not grouped.has_key(group):
                grouped[group] = []
            grouped[group].append(x)
        return grouped
        
        return self._cache.values()
                
    security.declareProtected(ManagePortal, 'setMaximumCacheAge')
    def setMaximumCacheAge(self, age):
        self.maximumCacheAge = age

    security.declareProtected(ManagePortal, 'setPeriod')
    def setPeriod(self, period):
        self.checkingPeriod = float(period)

    security.declareProtected(ManagePortal, 'setGracefulUpdateAge')
    def setGracefulUpdateAge(self, age):
        assert age < self.maximumCacheAge, \
            "Graceful update age is larger than maximum cache age."
        self.gracefulUpdateAge = age
     
    security.declareProtected(ManagePortal, 'setDefaultURIPrefix')
    def setDefaultURIPrefix(self, defaultURIPrefix):
        assert type(defaultURIPrefix) in StringTypes
        self.defaultURIPrefix = defaultURIPrefix
        
    security.declareProtected(ManagePortal, 'queueSize')
    def queueSize(self):
        global retrieval_queue
        return retrieval_queue.qsize()

    security.declareProtected(ManagePortal, 'workerThreadCount')
    def workerThreadCount(self):
        """Returns the number of worker threads currently running."""
        global thread_count
        return thread_count

    security.declareProtected(ManagePortal, 'activateWorkerThread')
    def activateWorkerThread(self):
        if not pool_threads.acquire(0):
            raise TooManyThreadsRunning
        global thread_count
        thread_count += 1
        thread = URLCheckerThread(retrieval_queue)
        thread.start()
        thread_manager.append(thread)

    security.declareProtected(ManagePortal, 'getThreadInformation')
    def getThreadInformation(self):
        global thread_manager
        info = {}
        names = [ x.getName() for x in thread_manager ]
        activity =  [ x.isAlive() for x in thread_manager ]
        info = zip(names, activity)
        return info

    security.declareProtected(ManagePortal, 'activateSchedulerThread')
    def activateSchedulerThread(self):
        global thread_queue, thread_manager, pool_scheduler, scheduler_count
        thread = LinkSchedulerThread(retrieval_queue, self.getPhysicalPath())
        thread.start()
        scheduler_count += 1
        thread_manager.append(thread)

    security.declareProtected(ManagePortal, 'workerThreadCount')
    def schedulerThreadCount(self):
        """Returns the number of worker threads currently running."""
        global scheduler_count
        return scheduler_count


    security.declareProtected(ManagePortal, 'countWorkingThreads')
    def countWorkingThreads(self):
        global thread_count
        return thread_count
        
    ### Private methods
    security.declarePrivate('__queueAddLink')
    def __queueAddLink(self, info):
        """Schedules a new check.

        Info must be a ILinkInfo object."""
        #zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Queue", "\tManually queueing %s\n" % info.url)
        retrieval_queue.put((self.getPhysicalPath(), info.getId())) 

    ### UI convenience methods
    security.declarePublic("isManaged")
    def isManaged(self, object):
        """Returns true/false for objects if they can be checked."""
        try:
           type = object.Type()
        except:
            return 0

        return type in self.getRetrievableTypes()

    security.declareProtected("View", "lastURLs")
    def lastURLs(self):
        """Return the last URLs that have been checked."""
        global last_urls
        return tuple(last_urls)

    security.declareProtected("View", "countLinks")
    def countLinks(self):
        """Return the number of links the linkchecker knows."""
        return len(self.objectIds(['Link Information']))

    security.declareProtected("View", "getLinksForObject")
    def getLinksForObject(self, object, nobrains=0):
        """Return a list of link information objects that are valid for this object."""
        sm = getSecurityManager()
        if not sm.checkPermission("View", object):
            raise "Unauthorized", "Can't check object for links"   
        links = self.link_catalog(objects=[self.getUIDForObject(object)])
        if nobrains:
            links = [ x.getObject() for x in links ]
        return links

    security.declareProtected("View", "getLinksForAuthenticatedMember")
    def getGroupedLinksForAuthenticatedMember(self):
        """Returns a list of link information objects that are valid for the authenticated user."""
        sm = getSecurityManager()
        user = sm.getUser()
        docs = self.portal_catalog(Creator=user.getId())
        uid = [ self.getUIDForObject(x.getObject()) for x in docs ]
        links = self.link_catalog(uid=uid)
        groups = {'red':[], 'orange':[], 'grey':[], 'green':[]}
        for link in links:
                groups[link.state].append(link)
                        
        return groups

    security.declareProtected("Manage portal", "getDocumentsInStateWithOwner")
    def getDocumentsInStateWithOwner(self, state="red"):
        links = self.link_catalog(state=[state,])
        links = [ x.getObject() for x in links ]
        links_ = []
        documents = {}
        for link in links:
            for doc in link.getObjects():
                key = doc.getPhysicalPath()
                if not documents.has_key(key):
                    # Add a new entry for this document
                    new = documents[key] = {}
                    new["document"] = doc
                    new["total"] = len(self.link_catalog(objects=[self.getUIDForObject(doc)]))
                    new["instate"] = 0
                    owner = doc.owner_info()["id"]
                    member = self.portal_membership.getMemberById(owner)
                    if member is None:
                        new["owner_mail"] = ""
                        new["owner"] = owner
                    else:
                        new["owner_mail"] = member.getProperty("email")
                        new["owner"] = member.getProperty("fullname", owner)
                    item = new
                else:
                    # Get the old entry
                    item = documents[key]

                item["instate"] += 1
        return documents.values()

    security.declareProtected("Manage portal", "getLinksInStateWithOwner")
    def getLinksInStateWithOwner(self, state="red"):
        links = self.link_catalog(state=[state,])
        links = [ x.getObject() for x in links ]
        links_ = []
        for link in links:
            for doc in link.getObjects():
                item = {}
                item["url"] = link.url
                item["reason"] = link.reason
                item["lastcheck"] = link.lastcheck
                item["document"] = doc
                owner = doc.owner_info()["id"]
                member = self.portal_membership.getMemberById(owner)
                if member is None:
                    item["owner_mail"] = ""
                    item["owner"] = owner
                else:
                    item["owner_mail"] = member.getProperty("email")
                    item["owner"] = member.getProperty("fullname", owner)
                links_.append(item)
        return links_

    security.declareProtected("Manage portal", "getBalancedScoreCard")
    def getBalancedScoreCard(self):
        status = {}

        # Total amount of links (distinct)
        status['total_distinct'] = len(self.objectValues(["Link Information"]))

        # Total amount of objects
        status['total_documents'] = len(self.link_catalog.Indexes['objects'].uniqueValues())

        # Documents per link
        values = []
        for x in self.objectValues(["Link Information"]):
            values.append(len(x.objects))
        values = values or [0]
        status['documents_per_link_min'] = min(values)
        status['documents_per_link_max'] = max(values)
        status['documents_per_link_avg'] = "%.02f" % reduce(lambda x,y:(x+y)/2.0, values)

        # Total amount of links
        status['total'] = reduce(lambda x,y:x+y, values)
        
        # Links per document
        documents = {}
        for x in self.objectValues(["Link Information"]):
            for ob in x.objects:
                count = documents.get(ob, 0)
                count += 1
                documents[ob] = count
        documents = documents.values() or [0]
        status['links_per_document_min'] = min(documents)
        status['links_per_document_max'] = max(documents)
        status['links_per_document_avg'] = "%.02f" % reduce(lambda x,y:(x+y)/2.0, documents)

        # Links per state (total)
        states = {'red':0, 'grey':0, 'green':0, 'orange':0}
        for x in self.objectValues(["Link Information"]):
            states[x.state] += 1
        for x in states.keys():
            status['total_%s_links' % x] = states[x]
            try:
                status['relative_%s_links' % x] = "%0.2f" % (float(states[x]) / status['total_distinct'] * 100)
            except ZeroDivisionError:
                status['relative_%s_links' % x] = "0.00"

        # Documents with state ...
        states = {'red': [], 'grey': [], 'green': [], 'orange': []}
        for x in self.objectValues(["Link Information"]):
            # We need to make it this complicated because we don't want to count documents
            # multiple times
            for y in x.objects:
                if y not in states[x.state]:
                    states[x.state].append(y)
        for x in states.keys():
            status['total_%s_documents' % x] = len(states[x])
            try:
                status['relative_%s_documents' % x] = "%.02f" % (float(len(states[x])) / status['total_documents'] * 100)
            except ZeroDivisionError:
                status['relative_%s_documents' % x] = "0.00"

        return status

    ##########################################################################
    # Mail notification methods

    security.declareProtected("Linkchecker: Run mail notifications", "runMailNotification")
    def runMailNotification(self):
        """Runs the mail notification code to create the mails
           for every user."""
        # setup hour detection
        hour = str(int(time.strftime("%H")))    # remove trailing 0 from the hour
        # get mailhost
        try:
            mailhost = self.MailHost
        except AttributeError:
            zLOG.LOG("CMFLinkChecker", zLOG.ERROR, "Mail notification", 
                     "\tCan't run notification: No Mailhost object reachable.")
            return

        # Count the mails for admin notification
        mail_count = 0
        for member in self.portal_membership.listMembers():
            # Check if the user can get a report at all:
            email = member.getProperty('email', '')
            if not email:
                continue

            if not member.getProperty('lc_notify_details', []):
                continue

            # Is this the right hour to send a report?
            frequency = member.getProperty('lc_notify_frequency', [])
            if not hour in frequency:
                continue

            report = {}
            seenlist = []

            for level in member.getProperty('lc_notify_details'):
                report[level] = []

            changes_only = member.getProperty('lc_notify_changes_only', 0)
            last_report = member.getProperty('lc_notify_last_notification', DateTime())

            # First gather all links
            for document in self.portal_catalog(Creator=member.getId()):
                links = self.getLinksForObject(document.getObject(), nobrains=1)
                for link in links:
                    if link.getId in seenlist:
                        continue
                    seenlist.append(link.getId)

                    # If we are only showing changes, go ahead.
                    if changes_only:
                        if link.lastupdate < last_report:
                            continue

                    if link.state in report.keys():
                        report[link.state].append(link)

                    if changes_only and link.laststatus in report.keys():
                        report[link.laststatus].append(link)

                    # Resolve object references
                    link.objects = tuple([ self.getObjectForUID(x) for x in link.objects])
      
            # Now get the template and create the html for the output
            template = self.portal_url.getPortalObject().lc_report_template

            keys = report.keys()
            states = [ x for x in ['red', 'orange', 'grey', 'green'] if x in keys]
            del keys
            
            mailbody = template(states=states, report=report, member=member, hour=hour)
            member.setMemberProperties({'lc_notify_last_notification': DateTime()})

            mMsg = "To: %s\n" % email
            mMsg = mMsg + "From: %s <%s>\n" % (self.email_from_name, self.email_from_address) # Portal configuration
            mMsg = mMsg + "Mime-Version: 1.0\n"
            mMsg = mMsg + "Content-Type: text/html; charset=%s\n\n" % self.portal_properties.site_properties.default_charset
            mMsg = mMsg + mailbody
            mSubj = "Link report for %s o'clock" % hour
            mailhost.send(mMsg, subject=mSubj, encode='base64')
            mail_count += 1

        return "Sent Mails: %s" % mail_count
    
InitializeClass(LinkCheckerTool)

#### Threading code.
# We only use one worker thread. Object references
# should be stable to work with.

# XXX This is somewhat whacky, but it works.

stopSignal = LinkInfo(url="")

class LinkSchedulerThread(Thread):
    """Schedules links in the site, and manages mailout triggers."""

    def __init__(self, queue, linkchecker):
        Thread.__init__(self, name="LinkScheduler")
        self.queue = queue
        self.linkchecker = '/'.join(linkchecker)
        self._stop = False

    def run(self):
        global pool_scheduler, scheduler_count
        if not pool_scheduler.acquire(0):
            raise TooManyThreadsRunning
        while not self._stop:
            # We don't want the scheduler to die ...
            try:
                self.run2()
            except:
                pass
        pool_threads.release()
        scheduler_count -= 1
            
    def run2(self):
        zLOG.LOG('CMFLinkChecker', 0, 'Initialized', '\tScheduling thread has been initialized.\n')

        # mailout setup
        last_notification_hour = None

        # calculate timing for scheduler
        check_app = None
        links = []
        linkid = None

        info = get_module_info("Zope")
        tm = info[7]
        root = info[2]()
        while 1:
            try:
                tm.begin()
                linkchecker = root.unrestrictedTraverse(self.linkchecker)
                print linkchecker

                if not links:   # Gather a new list of links and recalculate timings
                    period = linkchecker.checkingPeriod
                    links = linkchecker.objectIds(["Link Information"])
                    shuffle(links)
                    if len(links): # 0-division protection
                        pause = (60*60*period)/float(len(links))
                    else:
                        time.sleep(30)   # No links? No checks.
                        pause = 5
                        continue
                    zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Automatic Scheduler", "\tPause set to %s seconds\n" % pause)

                linkid = links.pop()
                link = getattr(linkchecker, linkid, None)
                if link is None:
                    zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Automatic Scheduler", "\tNot queuing stale link %s\n" % linkid)
                    
                zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Automatic Scheduler", "\tQueuing %s\n" % link.url)
                self.queue.put((linkchecker.getPhysicalPath(), link.getId()))
            except ReadConflictError:
                zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "ReadConflictError accessing %s. Rescheduling and pausing." % linkid)
                time.sleep(1)
                tm.abort()
                continue
            except:
                tm.abort()
            else:
                # We don't write anyway.
                tm.abort()
            time.sleep(pause)

class URLCheckerThread(Thread):

    def __init__(self, queue):
        Thread.__init__(self, name="URLChecker")
        self.queue = queue
        self._stop = False

    def stop(self):
        self._stop = True

    def run(self):
        while not self._stop:
            # We don't want the checker to die ...
            try:
                self.run2()
            except:
                ee, ev, et = sys.exc_info()

                info = get_module_info("Zope")
                tm = info[7]
                root = info[2]
                zLOG.LOG('CMFLinkChecker', 0, 'Dead thread', '\Thread died: %s:%s\n%s' % (ee,ev, '\n'.join(traceback.format_tb(et))))
                try:
                    error_log = root.unrestrictedTraverse('/error_log')
                    tm.begin()
                    if isinstance(error_log, SiteErrorLog):
                        error_log.raising((ee, ev, et))
                    tm.commit()
                except:
                    pass

        global pool_threads, thread_count
        pool_threads.release()
        thread_count -= 1

    def run2(self):
        zLOG.LOG('CMFLinkChecker', 0, 'Initialized', '\tWorker thread has been initialized.\n')
        info = get_module_info("Zope")
        tm = info[7]
        root = info[2]()
        while not self._stop:
            # Check if there is something in the queue
            try:
                linkchecker, key = self.queue.get(block=0)
            except Queue.Empty:
                time.sleep(1)
                continue

            try:
                tm.begin()

                # Get a handle to the linkchecker
                linkchecker = root.unrestrictedTraverse('/'.join(linkchecker))
                # Get the link to check
                try:
                    item = getattr(linkchecker, key)
                except AttributeError: # Ignore attribute errors, eventually the link is gone by now
                    tm.abort()
                    continue
                except ReadConflictError:
                    zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "ReadConflictError accessing %s. Rescheduling and pausing." % key)
                    self.queue.put((linkchecker, key))
                    time.sleep(1)
                    continue
                
                item.lastcheck = DateTime()
                zLOG.LOG("CMFLinkChecker", zLOG.BLATHER, "Checker", "\tChecking %s\n" % item.url)

                # Determine the plugin for checking code
                type, url = urllib.splittype(item.url)
                
                checker = protocols.registry.get(type, protocols.registry.get('unsupported'))
                try:
                    checker.check(item)
                    # Remember the last 5 URLs for UI purposes
                    global last_urls
                    if len(last_urls) > 5:
                        last_urls.pop()
                    last_urls.insert(0, item.url)
                except Exception, message:
                    zLOG.LOG('CMFLinkChecker', zLOG.ERROR, 'Worker Thread', 
                            'Checker for type %s on link %s had an error: %s' 
                            % (type, item.url, message))
                    # No need to save this transaction
                    tm.abort()
                    continue

                # Record metadata. I wish the TM would provide that as well
                get_transaction().note("Checking link %s" % item.url)
                get_transaction().setUser("CMFLinkChecker")
                tm.commit()

            except Exception, m:
                # Abort the transaction and re-raise
                tm.abort()
                raise   # Will be logged one level above
             

            # we are doing cooperative multitasking
            # this thread doesn't need too much power
            time.sleep(0.1)

        zLOG.LOG('CMFLinkChecker', zLOG.INFO, 'Stopped', '\tWorker thread stopped.\n')
        global thread_count
        thread_count -= 1
